#!/bin/bash

set -e

usage() {
	cat <<END_USAGE
Usage: $PROG COMMAND [OPTION...]

 Commands:
  -h, --help                  display this help message
  --checkout [MODULE]         checkout MODULE
  --update [MODULE]           update MODULE
  --prep [MODULE]             prepare MODULE
  --build [MODULE]            build MODULE
  --install [MODULE]          install MODULE
  --buildall [MODULE]         prepare, build, and install MODULE
  --version [MODULE]          get MODULE version
  --release [MODULE]          get MODULE release
  --moddep                    get rbh module file dependencies of MODULE
  --srcdep                    get source dependencies of MODULE
  --abi-major [DRIVER]        get ABI major version of XServer DRIVER
  --abi-minor [DRIVER]        get ABI minor version of XServer DRIVER
  --prefix                    get installation PREFIX (e.g. /usr/local)
  --libdir                    get object code libraries (e.g. PREFIX/lib)
  --mandir                    get man documentation (e.g. PREFIX/share/man)
  --sysconfdir                get man system configuration (e.g. /etc)
  --datadir                   get data files (e.g. PREFIX/share)
  --bindir                    get binary executable files (e.g. PREFIX/bin)
  --localstatedir             get local state directory (e.g. PREFIX/var)
  --includedir                get directory with C headers (e.g. PREFIX/include)
  --srcdir                    get directory where source is extracted
  --arch                      get arch of RBH_TARGET (e.g. i386/x86_64)
END_USAGE
}

if [[ -r "${RBH_CONF:=${0%/*}/rbh.conf}" ]]; then
	. "$RBH_CONF"
fi

RBH_ARGS=(
	"help"
	"checkout::"
	"update::"
	"prep::"
	"build::"
	"install::"
	"buildall::"
	"version:"
	"release:"
	"abi-major:"
	"abi-minor:"
	"prefix:"
	"libdir:"
	"mandir:"
	"sysconfdir:"
	"datadir:"
	"bindir:"
	"localstatedir:"
	"includedir:"
	"docdir:"
	"srcdir:"
	"moddep:"
	"srcdep:"
	"arch"
)

rbh_query_for_arch() {
	if command -v "rpm" >/dev/null; then
		echo $(rpm --eval %{_host})
	elif command -v "dpkg-architecture" >/dev/null; then
		echo $(dpkg-architecture -qDEB_HOST_GNU_TYPE)
	else
		echo $(uname -m)-unknown-linux-gnu
	fi
}

rbh_arch() {
	local arch=${RBH_TARGET%%-*}

	if [[ "$arch" = "i"?"86" ]]; then
		arch="i386"
	fi

	echo $arch
}

rbh_isabits() {
	if [[ "$(rbh_arch)" = "i386" ]]; then
		echo "32"
	else
		echo "64"
	fi
}

rbh_include_modules() {
	local modulefile
	for modulefile in "$@"; do
		if [ -f "$modulefile" ]; then
			. "$modulefile"
		fi
	done
}

: ${RBH_CC:=gcc}
: ${RBH_CXX:=g++}
: ${RBH_TARGET:=$(rbh_query_for_arch)}
: ${RBH_ISABITS:=$(rbh_isabits)}
: ${RBH_CFLAGS:="-m$RBH_ISABITS"}
: ${RBH_CXXFLAGS:="-m$RBH_ISABITS"}
: ${RBH_LDFLAGS:="-m$RBH_ISABITS"}
if command -v "cmake3" >/dev/null; then
: ${RBH_CMAKE:=cmake3}
else
: ${RBH_CMAKE:=cmake}
fi
if command -v "ninja-build" >/dev/null; then
: ${RBH_NINJA:=ninja-build}
else
: ${RBH_NINJA:=ninja}
fi
: ${RBH_ROOTCMD=sudo}
: ${RBH_BUILD_RELEASE:=1}
: ${RBH_BUILD_DIR:="${0%/*}/BUILD"}
: ${RBH_SOURCES_DIR:="${0%/*}/SOURCES"}
: ${RBH_DESTDIR:="/tmp"}
: ${RBH_SYSPREFIX:="/usr"}
: ${RBH_PREFIX:="/usr/local"}
: ${RBH_BINDIR:="$RBH_PREFIX/bin"}
: ${RBH_SYSLIBDIR:="$RBH_SYSPREFIX/lib"}
: ${RBH_LIBDIR:="$RBH_PREFIX/lib"}
: ${RBH_SYSCONFDIR:="$RBH_PREFIX/etc"}
: ${RBH_DATADIR:="$RBH_PREFIX/share"}
: ${RBH_DOCDIR:="$RBH_DATADIR/doc"}
: ${RBH_MANDIR:="$RBH_DATADIR/man"}
: ${RBH_LOCALSTATEDIR:="$RBH_PREFIX/var"}
: ${RBH_INCLUDEDIR:="$RBH_PREFIX/include"}
: ${RBH_JOBS:=$(nproc --ignore 2)}
: ${RBH_MODULE_DIR:="${0%/*}/*/rbh-modules"}
: ${RBH_MODULES:="
	kernel
	utils
	firmware
	libdrm
	llvm
	libvdpau
	libva
	vainfo
	mesa
	ddxati
	ddx
	gstomx
	ffmpeg
	mpv
	umr
"}
: ${RBH_FDO_GIT_BASE_URL:='http://anongit.freedesktop.org/git'}
: ${RBH_KDO_GIT_BASE_URL:='http://git.kernel.org/pub/scm/linux/kernel/git'}
: ${RBH_GITHUB_BASE_URL:='https://github.com'}

# Module core
: ${RBH_CORE_UNI_PREFIX:=$RBH_PREFIX}
: ${RBH_CORE_UNI_CREATE_TREE:="no"}

GIT_ARCHIVE_PREFIX="drivers"

# Import external modules
rbh_include_modules $RBH_MODULE_DIR/*.conf
rbh_include_modules $RBH_MODULE_DIR/*.rbh

RBH_TEMPDIR=

lnrs() {
	local p
	local i=1
	local s=$(readlink -e $1)
	local d=$(readlink -f $2)

	if [[ "$s" != "$d" ]]; then
		$RBH_ROOTCMD ln -rsf $s $d
	fi
}

rbh_setup_ccache() {
	if [[ -n "$RBH_CCACHE_PATH" ]]; then
		export PATH=${PATH//$RBH_CCACHE_PATH:}
		export PATH="$RBH_CCACHE_PATH:$PATH"
	else
		export CCACHE_DISABLE=1
	fi
}

rbh_init() {
	trap rbh_cleanup 0 1 2 15
	RBH_TEMPDIR=`mktemp -d -t rbh.XXXXXXXXXX`
	RBH_SOURCES_DIR=$(readlink -f $RBH_SOURCES_DIR)
	RBH_BUILD_DIR=$(readlink -f $RBH_BUILD_DIR)
	rbh_setup_ccache
	if [[ "$RBH_WAYLAND_SUPPORT" == "yes" ]]; then
		export BUILD_WAYLAND_SUPPORT=1
	else
		export BUILD_WAYLAND_SUPPORT=0
	fi
}

rbh_cleanup() {
	[[ -z "$RBH_TEMPDIR" ]] || rm -rf "$RBH_TEMPDIR" || true
}

rbh_cflags() {
	local flags
	local module=$1

	eval flags=\$RBH_${module^^}_CFLAGS
	echo ${flags:-$RBH_CFLAGS}
}

rbh_cxxflags() {
	local flags
	local module=$1

	eval flags=\$RBH_${module^^}_CXXFLAGS
	echo ${flags:-$RBH_CXXFLAGS}
}

rbh_ldflags() {
	local flags
	local module=$1

	eval flags=\$RBH_${module^^}_LDFLAGS
	echo ${flags:-$RBH_LDFLAGS}
}

rbh_prefix() {
	local prefix
	local module=$1

	eval prefix=\$RBH_${module^^}_PREFIX
	echo ${prefix:-$RBH_PREFIX}
}

rbh_moddep() {
	local moddep
	local module=$1

	eval moddep=\$RBH_${module^^}_MODDEP
	echo ${moddep}
}

rbh_srcdep() {
	local srcdep
	local module=$1

	eval srcdep=\$RBH_${module^^}_SRCDEP
	echo ${srcdep}
}

rbh_libdir() {
	local module=$1
	eval local dir=\$RBH_${module^^}_LIBDIR

	echo ${dir:-${RBH_LIBDIR/$RBH_PREFIX/$(rbh_prefix $module)}}
}

rbh_mandir() {
	local module=$1
	eval local dir=\$RBH_${module^^}_MANDIR

	echo ${dir:-${RBH_MANDIR/$RBH_PREFIX/$(rbh_prefix $module)}}
}

rbh_sysconfdir() {
	local module=$1
	eval local dir=\$RBH_${module^^}_SYSCONFDIR

	echo ${dir:-${RBH_SYSCONFDIR/$RBH_PREFIX/$(rbh_prefix $module)}}
}

rbh_datadir() {
	local module=$1
	eval local dir=\$RBH_${module^^}_DATADIR

	echo ${dir:-${RBH_DATADIR/$RBH_PREFIX/$(rbh_prefix $module)}}
}

rbh_docdir() {
	local module=$1
	eval local dir=\$RBH_${module^^}_DOCDIR

	echo ${dir:-${RBH_DOCDIR/$RBH_PREFIX/$(rbh_prefix $module)}}
}

rbh_bindir() {
	local module=$1
	eval local dir=\$RBH_${module^^}_BINDIR

	echo ${dir:-${RBH_BINDIR/$RBH_PREFIX/$(rbh_prefix $module)}}
}

rbh_localstatedir() {
	local module=$1
	eval local dir=\$RBH_${module^^}_LOCALSTATEDIR

	echo ${dir:-${RBH_LOCALSTATEDIR/$RBH_PREFIX/$(rbh_prefix $module)}}
}

rbh_includedir() {
	local module=$1
	eval local dir=\$RBH_${module^^}_INCLUDEDIR

	echo ${dir:-${RBH_INCLUDEDIR/$RBH_PREFIX/$(rbh_prefix $module)}}
}

rbh_srcdir() {
	local module=$1
	eval local src_url=\$RBH_${module^^}_GIT_URL
	eval local src_prefix=\$GIT_ARCHIVE_${module^^}_PREFIX
	local dir=${src_url##*/}

	echo $src_prefix${dir%.git}
}

rbh_pkg_config_path() {
	local module=$1 subdir=$2
	local fn="rbh_pkg_config_${module}_path"
	local lib data
	local syslib="$RBH_SYSLIBDIR/pkgconfig"

	if command -v "$fn" >/dev/null; then
		$fn $module $subdir ||:
	else
		lib=$(rbh_libdir $module)/pkgconfig
		data=$(rbh_datadir $module)/pkgconfig
		export PKG_CONFIG_PATH="$lib:$data:$syslib"
	fi
}

rbh_chrpath() {
	local i path=$1

	for i in $(find $path -executable -a -not -type d); do
		file $i | awk -F: '$2~/ELF/ {print $1}' |
			xargs $RBH_ROOTCMD chrpath -d
	done
}

rbh_get_url_subdir() {
	local module=$1
	local url_param=$2
	local subdir_param=$3
	local src_url dst_dir

	eval src_url=\$RBH_${module^^}_GIT_URL
	if [[ -z $src_url ]]; then
		eval $url_param=''
		eval $subdir_param=''
		echo "Warning: URL is not set up for the $module module"
	else
		eval $url_param=\$src_url

		eval dst_dir=\$RBH_${module^^}_MODULE_DIR
		if [[ -z $dst_dir ]]; then
			dst_dir=${src_url##*/}
			dst_dir=${dst_dir%.git}
		fi
		eval $subdir_param=\$dst_dir
	fi
}

rbh_checkout() {
	# TODO: check syntax
	local module url dir subdir patches commit hook
	local ver prefix

	for module in ${1:-$RBH_MODULES}; do
		rbh_get_url_subdir ${module} url subdir
		[[ -z $url ]] && continue

		dir="$RBH_SOURCES_DIR/$subdir"
		eval commit=\$RBH_${module^^}_COMMIT

		if [[ ! -d "$dir/.git" ]]; then
			rbh_log "checkout"
			git clone $url $dir
			patches="${0%/*}/${module}-patches.tar.xz"

			if [[ -e "$patches" ]]; then
				(cd  $RBH_SOURCES_DIR && tar Jxvf $patches)
			fi

			pushd $dir >/dev/null
			git checkout -b rbh ${commit:-master}

			if [[ -e "$patches" ]]; then
				git am --whitespace=nowarn ../${module}-patches/*.patch
			fi

			popd >/dev/null
		fi

		hook="post_checkout_${module}_hook"

		if command -v "$hook" >/dev/null; then
			$hook "$module" "$dir" ||:
		fi

		rbh_archive "$module" "$dir"
	done
}

rbh_archive() {
	local module=$1
	local dir=$2
	local tarext="tar.gz"
	local prefix subdir ver hook m

	eval prefix=\$GIT_ARCHIVE_${module^^}_PREFIX
	eval subdir=\$GIT_ARCHIVE_${module^^}_SUBDIR
	ver=$(rbh_version $module)

	if [[ "$module" = *"_uni" ]]; then
		m=${module%_*}
	else
		m=$module
	fi

	pushd $dir >/dev/null
	if [[ "$GIT_ARCHIVE_NO_COMPRESS" == "yes"  ]]; then
		git archive --format=tar \
			--prefix="${prefix}${dir##*/}/"	\
			HEAD ${subdir:+-- $subdir} \
			>$RBH_SOURCES_DIR/$m-$ver.tar
		tarext="tar"
	else
		git archive --format=tar \
			--prefix="${prefix}${dir##*/}/"	\
			HEAD ${subdir:+-- $subdir} \
			| gzip >$RBH_SOURCES_DIR/$m-$ver.tar.gz
	fi


	git submodule update --init --recursive
	git submodule foreach "git archive --format=$tarext --prefix=${dir##*/}/\$path/ HEAD \
		>$RBH_SOURCES_DIR/\$name.$tarext"
	popd >/dev/null

	hook="rbh_archive_${module}_hook"

	if command -v "$hook" >/dev/null; then
		$hook "$module" "$dir" "$ver" ||:
	fi
}

rbh_archive_addfiles() {
	local files=$1
	local prefix=$2
	local module=${3%_uni}
	local ver=$4

	[[ "$GIT_ARCHIVE_NO_COMPRESS" == "yes"  ]] || \
		gunzip $RBH_SOURCES_DIR/$module-$ver.tar.gz
	tar --transform "s|^|$prefix/|" --exclude-vcs \
		-rf $RBH_SOURCES_DIR/$module-$ver.tar $files
	[[ "$GIT_ARCHIVE_NO_COMPRESS" == "yes"  ]] || \
		gzip $RBH_SOURCES_DIR/$module-$ver.tar
}

rbh_update() {
	# TODO: check syntax
	local module url dir subdir

	for module in ${1:-$RBH_MODULES}; do
		eval url=\$RBH_${module^^}_GIT_URL
		subdir=${url##*/}
		subdir=${subdir%.git}
		dir="$RBH_SOURCES_DIR/$subdir"
		rbh_log "update"
		pushd $dir >/dev/null
		git pull
		popd >/dev/null
	done
}

rbh_prep() {
	# TODO: check syntax
	local module url subdir prep

	for module in ${1:-$RBH_MODULES}; do
		eval url=\$RBH_${module^^}_GIT_URL
		subdir=${url##*/}
		subdir=${subdir%.git}
		buildsubdir="${RBH_BUILD_SUBDIR:-$subdir}"
		rbh_log "prep"
		prep="rbh_prep_${module}"
		mkdir -p "$RBH_BUILD_DIR/$buildsubdir"

		if command -v "$prep" >/dev/null; then
			pushd $RBH_BUILD_DIR/$buildsubdir >/dev/null
			$prep "$module" "$subdir" "$buildsubdir"
			popd >/dev/null
		fi
	done
}

rbh_version() {
	local module subdir modver

	module=$1
	modver="rbh_version_${module}"
	subdir=$RBH_SOURCES_DIR/$(rbh_srcdir $module)

	if [[ -d "$subdir" ]] && command -v "$modver" >/dev/null; then
		pushd $subdir >/dev/null
		rbh_version_${module}
		popd >/dev/null
	else
		rbh_version_default
	fi
}

rbh_version_xserver() {
	mkdir -p m4
	autoreconf --force --install >/dev/null 2>&1 || true
	./configure --version | awk 'NR == 1 {print $NF}'
}

rbh_version_default() {
	if [[ -n "$RBH_BUILD_VERSION" ]]; then
		echo $RBH_BUILD_VERSION
	else
		echo "$(date +%y).$(( ($(date +%m) - 1) / 3 + 1 ))0.99"
	fi
}

rbh_release() {
	echo ${RBH_BUILD_RELEASE}${RBH_DIST_TAG}
}

rbh_build() {
	# TODO: check syntax
	local module url subdir

	for module in ${1:-$RBH_MODULES}; do
		eval url=\$RBH_${module^^}_GIT_URL
		subdir=${url##*/}
		subdir=${subdir%.git}
		buildsubdir="${RBH_BUILD_SUBDIR:-$subdir}"
		rbh_log "build"
		rbh_build_$module "$module" "$subdir" "$buildsubdir"
	done
}

rbh_install() {
	# TODO: check syntax
	local module url subdir destdir install_args inst script

	for module in ${1:-$RBH_MODULES}; do
		eval url=\$RBH_${module^^}_GIT_URL
		subdir=${url##*/}
		subdir=${subdir%.git}
		buildsubdir="${RBH_BUILD_SUBDIR:-$subdir}"
		destdir=${RBH_DESTDIR:+" DESTDIR=$RBH_DESTDIR"}
		eval install_args=\$RBH_${module^^}_INSTALL_ARGS\$destdir
		rbh_log "install"
		script="rbh_install_script_${module}"
		inst="rbh_install_${module}"

		pushd $RBH_BUILD_DIR/$buildsubdir >/dev/null
		if command -v "$inst" >/dev/null; then
			$inst "$module" "$subdir" "$buildsubdir"
		else
			$RBH_ROOTCMD make install $install_args
			if command -v "$script" >/dev/null; then
				$script "$module" "$subdir" "$buildsubdir" "$install_args"
			fi
		fi
		popd >/dev/null
	done
}

rbh_buildall() {
	# TODO: check syntax
	local module url subdir phase

	for module in ${1:-$RBH_MODULES}; do
		eval url=\$RBH_${module^^}_GIT_URL
		subdir=${url##*/}
		subdir=${subdir%.git}

		for phase in prep build install; do
			rbh_$phase "$module"
		done
	done
}

rbh_install_core_uni() {
	local module=$1
	local bindir=$(rbh_bindir $module)
	local prefixdir=$(rbh_prefix $module)
	local datadir=$(rbh_datadir $module)
	local mandir=$(rbh_mandir $module)
	local docdir=$(rbh_docdir $module)

	if [[ "$RBH_CORE_UNI_CREATE_TREE" = "yes" ]]; then
		mkdir -p ${RBH_DESTDIR}$bindir/
		mkdir -p ${RBH_DESTDIR}$prefixdir/{lib,lib64}/pkgconfig/
		mkdir -p ${RBH_DESTDIR}$datadir/
		mkdir -p ${RBH_DESTDIR}$mandir/man{3,4,7}/
		mkdir -p ${RBH_DESTDIR}$prefixdir/etc
		mkdir -p ${RBH_DESTDIR}$docdir/
	fi
}

contain() {
	local e
	for e in "${@:2}"; do
		if [[ "$e" == "$1" ]]; then
			return 0
		fi
	done
	return 1
}

rbh_abi-major() {
	local dir drv=${1^^}

	dir=${RBH_XSERVER_GIT_URL##*/}
	dir=$RBH_SOURCES_DIR/${dir%.git}
	awk -v var=$drv -F'[ \t,()]+' '$1 ~ /^#define/ && $2 == "ABI_"var"_VERSION" { print $4 }' \
		$dir/hw/xfree86/common/xf86Module.h
}

rbh_abi-minor() {
	local dir drv=${1^^}

	dir=${RBH_XSERVER_GIT_URL##*/}
	dir=$RBH_SOURCES_DIR/${dir%.git}
	awk -v var=$drv -F'[ \t,()]+' '$1 ~ /^#define/ && $2 == "ABI_"var"_VERSION" { print $5 }' \
		$dir/hw/xfree86/common/xf86Module.h
}

rbh_build_generic_driver() {
	local module=$1 subdir=$2 buildsubdir=$3

	rbh_pkg_config_path $module $subdir
	cd $RBH_BUILD_DIR/$buildsubdir
	$RBH_SOURCES_DIR/$subdir/autogen.sh \
		--host=$RBH_TARGET \
		--prefix=$RBH_PREFIX \
		--libdir=$RBH_LIBDIR \
		--disable-static \
		--disable-silent-rules \
		CFLAGS="$(rbh_cflags $module)" \
		LDFLAGS="$(rbh_ldflags $module)"
	make -j$RBH_JOBS CC="$RBH_CC"
}

xver_major_minor() {
	local xver="$RBH_TEMPDIR/xorg-version"

	echo -e `\
		`"#include <stdio.h>\n"`
		`"#include <xorg/xorg-server.h>\n"`
		`"int main() {\n"`
		`"    int major=XORG_VERSION_CURRENT/10000000;\n"`
		`"    int minor=XORG_VERSION_CURRENT/100000%100;\n"`
		`"    printf(\"%d.%d\\\n\", major, minor);\n"`
		`"    return 0;\n"`
		`"}\n" \
	| $RBH_CC -o $xver -xc - && $xver
}

rbh_log() {
	echo "*** $(date +"%F %T") $module:" "$@"
}

format_args() {
	local i
	local delim=$1
	local count=0

	for i in ${RBH_ARGS[@]}; do
		if (( count == ${#RBH_ARGS[*]} - 1 )); then
			delim=""
		fi

		printf "%s%s" $i $delim
		(( count += 1 ))
	done
}

PROG=${0##*/}

if [[ "$#" -eq 0 ]]; then
	echo "Try '$PROG --help' for more information." >&2
	exit 1
fi

OPTS=$(getopt -o h -l $(format_args ",") -- "$@")
eval set -- "$OPTS"
rbh_init

while true; do
	case "$1" in
	-h|--help)
		usage
		exit 0
		;;
	--checkout|--update|--prep|--build|--install|--buildall)
		ACTION=${1#--}
		shift 2
		;;
	--prefix|--libdir|--mandir|--sysconfdir|--datadir|--bindir| \
	--localstate|--includedir|--docdir|--srcdir|--moddep|--srcdep| \
	--version|--release|--abi-major|--abi-minor|--arch)
		rbh_${1#--} $2
		exit $?
		;;
	--)
		shift
		break
		;;
	esac
done

rbh_$ACTION "$@"
exit $?
